
package pwrace2;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;

/**
 * This probram is a simple round robin calculator for multiple lane races and
 * multiple races per lane.  For example a track may have 4 lanes and may race
 * two times in each lane.  This program will calculate the optimum number of
 * participants, and the starting lineup so in each lane, the car number will
 * be incremented by 1 and each racer will race every other racer.
 *
 * This version of the calculater is different from the others in that the
 * previous version took a brute forse approach and incremented to the next
 * racer regardless if it was already used.  They then checked if it was a
 * valid linup.  The previous versions checked if it could skip a bunch, but
 * still took quite a bit of time to increment through all the positions.
 *
 * This version takes a much more mathematical or logical approach.
 * A position will select the next racer based on what is available
 * (or handed to it).  It then will check if it is a valid racer for that
 * position and what other racers get taken also, based on incrementing through
 * the lanes.  If it is, it creates another list of available racers minus what
 * was used and passes it on to the next position.  If no valid racer is found
 * for a position, it reports back to the previous position that it failed and
 * the previous position will increment and provide an updated list of available
 * racers.  This is more of a recursive algorithm and is the fastest.
 *
 * @author ghewitt
 */
public class RRobinPanel extends JPanel {
    // panel fields
    private JTextField numberOfLanesTextField;
    private JTextField racesPerLaneTextField;
    private JTextField idealNumOfRacersTextField;
    private JTextField addRacersTextField;
    private JTextField numberOfRacersTextField;
    private JTable resultsTable;
    private JButton calculateButton;
    private JButton stopButton;
    private JButton applyButton;
    private ResultsTableModel resultsTableModel;
    private JScrollPane resultsScrollPane;
    private JCheckBox useShortCutCheckBox;
    private JRadioButton displayAllRadioButton;
    private JRadioButton displayRowOneRadioButton;
    private JRadioButton displayThreeRadioButton;
    private JRadioButton displayNoneRadioButton;
    private ButtonGroup displayButtonGroup;
    private int numOfRacers;
    private int numOfLanes;
    private int racesPerLane;
    private int addRacers;
    private int startLineUp[][];
    private int availableLineUp[][];
    private boolean validLineUp;
    private Thread calculateRRThread;
    private CalculateRoundRobin calcRR;

    public RRobinPanel() {
        setLayout(new GridLayout(1,1));
        
        setLayout(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();

        validLineUp = false;
        numOfRacers = 0;
        numOfLanes = 4;
        racesPerLane = 1;
        addRacers = 0;
        availableLineUp = null;
        startLineUp = null;

        c.gridx = 0;
        c.gridy = 0;
        c.insets = new Insets(2,5,3,5);
        c.anchor = GridBagConstraints.LINE_START;
        c.gridwidth = 2;
        add(new JLabel("Number of Lanes:"), c);
        c.gridy = 1;
        add(new JLabel("Races Per Lane:"), c);
        c.gridy = 2;
        add(new JLabel("Ideal Number of Racers:"), c);
        c.gridy = 3;
        add(new JLabel("Additional Racers Above Ideal:"), c);
        c.gridwidth = 1;

        numberOfLanesTextField = new JTextField(3);
        numberOfLanesTextField.setText("" + numOfLanes);
        numberOfLanesTextField.addFocusListener(new LanesFocusListener());
        c.gridx = 2;
        c.gridy = 0;
        add(numberOfLanesTextField, c);

        racesPerLaneTextField = new JTextField(3);
        racesPerLaneTextField.setText("" + racesPerLane);
        racesPerLaneTextField.addFocusListener(new RacesFocusListener());
        c.gridy = 1;
        add(racesPerLaneTextField, c);

        idealNumOfRacersTextField = new JTextField(3);
        idealNumOfRacersTextField.setEditable(false);
        idealNumOfRacersTextField.setFocusable(false);
        idealNumOfRacersTextField.setText("13");
        c.gridy = 2;
        add(idealNumOfRacersTextField, c);

        addRacersTextField = new JTextField(3);
        addRacersTextField.setText("" + addRacers);
        c.gridy = 3;
        add(addRacersTextField, c);

        useShortCutCheckBox = new JCheckBox("Use B1 Short Cut");
        c.gridwidth = 2;
        c.gridx = 0;
        c.gridy = 4;
        add(useShortCutCheckBox,c);
        c.gridwidth = 1;
        
        calculateButton = new JButton("Calculate");
        calculateButton.addActionListener(new CalculateActionListener());
        c.gridx = 0;
        c.gridy = 5;
        add(calculateButton, c);

        resultsTableModel = new ResultsTableModel();
        resultsTable = new JTable(resultsTableModel);
        resultsScrollPane = new JScrollPane(resultsTable);
        resultsScrollPane.setPreferredSize(new Dimension(350, 100));
        resultsScrollPane.setBorder(BorderFactory.createTitledBorder("Results:"));
        resultsTable.setEnabled(false);
        resultsTable.setFocusable(false);
        c.gridx = 0;
        c.gridy = 6;
        c.gridwidth = 4;
        c.fill = GridBagConstraints.BOTH;
        c.weighty = 1;
        add(resultsScrollPane, c);
        c.fill = GridBagConstraints.NONE;
        c.weighty = 0;

        c.gridy = 7;
        c.gridwidth = 2;
        add(new JLabel("Number of Racers Needed:"), c);

        numberOfRacersTextField = new JTextField(3);
        numberOfRacersTextField.setEditable(false);
        numberOfRacersTextField.setFocusable(false);
        c.gridx = 2;
        c.gridwidth = 1;
        add(numberOfRacersTextField, c);

        c.gridx = 1;
        c.gridy = 5;
        stopButton = new JButton("Stop");
        stopButton.addActionListener(new stopListener());
        stopButton.setEnabled(false);
        add(stopButton, c);

        c.gridx = 3;
        applyButton = new JButton("Apply");
        applyButton.addActionListener(new applyListener());
        applyButton.setEnabled(false);
        add(applyButton, c);
        
        c.gridx = 3;
        c.gridy = 0;
        c.weightx = 1;
        add(new JLabel("Display Calculations"), c);
        c.gridy = 1;
        displayAllRadioButton = new JRadioButton("All");
        displayAllRadioButton.setFocusable(false);
        add(displayAllRadioButton, c);
        
        c.gridy = 2;
        displayRowOneRadioButton = new JRadioButton("First Row ");
        displayRowOneRadioButton.setSelected(true);
        displayRowOneRadioButton.setFocusable(false);
        add(displayRowOneRadioButton, c);

        c.gridy = 3;
        displayThreeRadioButton = new JRadioButton("First Three cells");
        displayThreeRadioButton.setFocusable(false);
        add(displayThreeRadioButton, c);

        c.gridy = 4;
        displayNoneRadioButton = new JRadioButton("None");
        displayNoneRadioButton.setFocusable(false);
        add(displayNoneRadioButton, c);
        
        displayButtonGroup = new ButtonGroup();
        displayButtonGroup.add(displayAllRadioButton);
        displayButtonGroup.add(displayRowOneRadioButton);
        displayButtonGroup.add(displayThreeRadioButton);
        displayButtonGroup.add(displayNoneRadioButton);
    }


    private class CalculateRoundRobin implements Runnable  {
        private boolean running;
        private boolean racerTest[];
        public void run()  {
            // actual algorithm for calculating the round robin heat line ups
            validLineUp = false;
            running = true;

            numOfRacers = numOfLanes * (numOfLanes-1) * racesPerLane + 1 + addRacers;
            numberOfRacersTextField.setText(""+ numOfRacers);
            numberOfRacersTextField.setBackground(getBackground());

            startLineUp = new int[racesPerLane][numOfLanes];

            Thread.yield();

            // setup positions and initialize;
            Position[][] incPositions =new Position[racesPerLane][numOfLanes-1];
            Position tempPos = null;
            for(int row = racesPerLane-1; row >= 0; row--)  {
                for(int col = numOfLanes-2; col >= 0; col--)  {
                    incPositions[row][col] = new Position(row, col+1, tempPos);
                    tempPos = incPositions[row][col];
                }
            }
            if(displayAllRadioButton.isSelected())  {
                for (int row = 0; row < incPositions.length; row++)  {
                    for (int col = 0; col< incPositions[0].length; col++)  {
                        incPositions[row][col].setDisplay(true);
                    }
                }
            }
            
            if(displayRowOneRadioButton.isSelected())  {
                for (int col = 0; col < incPositions[0].length; col++)  {
                    incPositions[0][col].setDisplay(true);
                }
            }
            
            if(displayThreeRadioButton.isSelected())  {
                for (int col = 0; col < 2 && col < incPositions[0].length; col++)  {
                    incPositions[0][col].setDisplay(true);
                }
            }


            // Loop untill startLineUp is valid
            while(!validLineUp && running)  {
                validLineUp = true;

                // initialize display table and starting lineup
                for (int row = 0; row < racesPerLane; row++)
                    for (int col = 0; col < numOfLanes; col++)
                        startLineUp[row][col] = 0;
                resultsTableModel.setData(startLineUp);

                // create and clears test array
                racerTest = new boolean[numOfRacers];
                racerTest[0] = true;
                 for (int row =1; row < numOfRacers; row ++)  {
                     racerTest[row] = false;
                 }

                // set increment position array sizes
                for(int row = 0; row < incPositions.length; row++)  {
                    for(int col = 0; col < incPositions[0].length; col++)  {
                        incPositions[row][col].setMaxRacers(numOfRacers);
                    }
                }

                // use short cut
                if(useShortCutCheckBox.isSelected())  {
                    // create initial available lineup
                    startLineUp[0][1] = 1;
                    resultsTableModel.setValueAt(1, 0, 1);
                    for (int i=1; i<numOfRacers-1; i++)  {
                        incPositions[0][1].addRacer(i);
                    }
                    // call first postion to begin calculations
                    validLineUp = incPositions[0][1].calcPosition();
                }
                else  {
                    // create initial available lineup
                    for (int i=1; i<numOfRacers; i++)  {
                        incPositions[0][0].addRacer(i);
                    }
                    // call first postion to begin calculations
                    validLineUp = incPositions[0][0].calcPosition();
                }

                // if not a valid lineup, increments and resize of arrays
                if (!validLineUp && running)  {
                    numOfRacers ++;
                    numberOfRacersTextField.setText(""+ numOfRacers);
                    numberOfRacersTextField.setBackground(Color.YELLOW);
                }
            }

            // display results, prepair for next input
            resultsTableModel.setData(startLineUp);
            calculateButton.setEnabled(true);
            numberOfLanesTextField.setEnabled(true);
            racesPerLaneTextField.setEnabled(true);
            addRacersTextField.setEnabled(true);
            useShortCutCheckBox.setEnabled(true);
            displayAllRadioButton.setEnabled(true);
            displayRowOneRadioButton.setEnabled(true);
            displayThreeRadioButton.setEnabled(true);
            displayNoneRadioButton.setEnabled(true);
            applyButton.setEnabled(true);
            stopButton.setEnabled(false);
        }

        public void stopCalc()  {
            running = false;
        }

        // Internal private class for each possition.
        // positions will only check current row up to themselves, no need to
        // go beyond or check other rows.

        private class Position  {
            private Position nextPosition;
            private int positionRow, positionCol, racerCount;
            private int[] availableRacers;
            private boolean isNext, display;


            public Position(int row, int col, Position next)  {
                nextPosition = next;
                positionRow = row;
                positionCol = col;
                availableRacers = null;
                racerCount = 0;
                isNext = true;
                display = false;
                if(nextPosition == null)  { isNext = false;  }
            }
            public void setDisplay(boolean disp)  {
                display = disp;
            }
            
            public void setMaxRacers(int mracer)  {
                availableRacers = new int[mracer];
                racerCount = 0;
            }

            public void addRacer(int racer)  {
                availableRacers[racerCount] = racer;
                racerCount++;
            }

            public boolean calcPosition()  {
                boolean validPosition = false;
                int index=0;
                int currentRacer, checkRacer;

                while(!validPosition && index < racerCount && running)  {
                    // select racer number;
                    validPosition = true;
                    currentRacer = availableRacers[index];

                    // check if valid
                    for(int col=0; col<positionCol && validPosition; col++)  {
                        // check if positionCol is zero column
                        checkRacer = startLineUp[positionRow][col] - currentRacer;
                        if(checkRacer <0) checkRacer += numOfRacers;
                        if(racerTest[checkRacer])  {
                            validPosition = false;
                        }
                        else  {
                            racerTest[checkRacer] = true;
                        }
                        // check if col is zero column
                        checkRacer = currentRacer - startLineUp[positionRow][col];
                        if(checkRacer <0) { checkRacer += numOfRacers; }
                        if(racerTest[checkRacer])  {
                            validPosition = false;
                        }
                        else  {
                            racerTest[checkRacer] = true;
                        }
                    }
                    // if valid create available lineup and call next position
                    if(validPosition)  {
                        startLineUp[positionRow][positionCol] = currentRacer;
                        if (display) {
                            resultsTableModel.setValueAt(currentRacer, positionRow, positionCol);

                        }
                        if (isNext)  {
                            for (int i = 0; i < racerCount; i++) {
                                if (!racerTest[availableRacers[i]]) {
                                    nextPosition.addRacer(availableRacers[i]);
                                }
                            }
                            validPosition = nextPosition.calcPosition();
                        }
                    }
                    else  { // if not valid, clear checks, increment index;
                        index++;
                        for (int i=0; i<racerCount; i++)  {
                            racerTest[availableRacers[i]] = false;
                        }
                    }
                }
                racerCount = 0;
                return validPosition;
            }
        }
    }

    private class ResultsTableModel extends AbstractTableModel  {
        private Integer data[][];

        public ResultsTableModel() {
            data = new Integer[1][4];

            for(int row=0; row<1; row++)
                for (int col=0; col<4; col++)
                    data[row][col] = new Integer(0);

        }
        
        synchronized public void setData(int[][] newData)  {
            if(newData[0].length != data[0].length)  {
                data = new Integer[newData.length][newData[0].length];
                fireTableStructureChanged();
            }
            if(newData.length != data.length)
                data = new Integer[newData.length][newData[0].length];

            for(int row=0; row< newData.length; row++)
                for(int col=0; col<newData[0].length; col++)
                    data[row][col] = new Integer(newData[row][col]+1);

            fireTableDataChanged();
        }

        synchronized public int getRowCount() {
            return data.length;
        }

        synchronized public int getColumnCount() {
            return data[0].length;
        }

        synchronized public Object getValueAt(int rowIndex, int columnIndex) {
            return data[rowIndex][columnIndex];
        }
        synchronized public void setValueAt(Integer value, int row, int col)  {
            data[row][col] = (Integer) (value+1);
            fireTableCellUpdated(row, col);
        }
    }

    private class CalculateActionListener implements ActionListener  {

        public void actionPerformed(ActionEvent e) {

            // disables everything untill finished
            numberOfRacersTextField.setBackground(getBackground());
            calculateButton.setEnabled(false);
            numberOfLanesTextField.setEnabled(false);
            racesPerLaneTextField.setEnabled(false);
            useShortCutCheckBox.setEnabled(false);
            addRacersTextField.setEnabled(false);
            displayAllRadioButton.setEnabled(false);
            displayRowOneRadioButton.setEnabled(false);
            displayThreeRadioButton.setEnabled(false);
            displayNoneRadioButton.setEnabled(false);
            applyButton.setEnabled(false);
            stopButton.setEnabled(true);
            stopButton.setBackground(calculateButton.getBackground());

            // verifies the inputs and starts calculations.
            boolean validInput = true;
            int lanes = 0, rpl = 0, addr = 0;
            try  {
                lanes = Integer.parseInt(numberOfLanesTextField.getText());
                if(lanes < 2) throw new Exception();
            }
            catch(Exception err)  {
                validInput = false;
                numberOfLanesTextField.setText("" + numOfLanes);
            }

            try  {
                rpl = Integer.parseInt(racesPerLaneTextField.getText());
                if(rpl < 1) throw new Exception();
            }
            catch(Exception err)  {
                validInput = false;
                racesPerLaneTextField.setText("" + racesPerLane);
            }
            try  {
                addr = Integer.parseInt(addRacersTextField.getText());
                if(addr < 0) throw new Exception();
            }
            catch(Exception err)  {
                validInput = false;
                addRacersTextField.setText("" + addRacers);
            }

            if (validInput)  {
                numOfLanes = lanes;
                racesPerLane = rpl;
                addRacers = addr;
                calcRR = new CalculateRoundRobin();
                calculateRRThread = new Thread(calcRR);
                calculateRRThread.start();
            }
            else  {
                calculateButton.setEnabled(true);
                numberOfLanesTextField.setEnabled(true);
                racesPerLaneTextField.setEnabled(true);
                useShortCutCheckBox.setEnabled(true);
                addRacersTextField.setEnabled(true);
                displayAllRadioButton.setEnabled(true);
                displayRowOneRadioButton.setEnabled(true);
                displayThreeRadioButton.setEnabled(true);
                displayNoneRadioButton.setEnabled(true);
                stopButton.setEnabled(false);

            }
        }
    }
    private class stopListener implements ActionListener  {

        public void actionPerformed(ActionEvent e) {
            stopButton.setBackground(Color.RED);
            calcRR.stopCalc();
        }
    }
    private class applyListener implements ActionListener  {

        public void actionPerformed(ActionEvent e) {
            applyButton.setEnabled(false);
            if(startLineUp !=null)  {
                availableLineUp = new int[startLineUp.length][startLineUp[0].length];
                for (int row = 0; row < startLineUp.length; row++)  {
                    for (int col = 0; col< startLineUp[0].length; col ++)  {
                        availableLineUp[row][col] = startLineUp[row][col];
                    }
                }
            }
        }
    }

    private class LanesFocusListener implements FocusListener {

        public void focusGained(FocusEvent e) {
            // not used
        }

        public void focusLost(FocusEvent e) {
            int lanes;
            try  {
                lanes = Integer.parseInt(numberOfLanesTextField.getText());
                if(lanes < 2) throw new Exception();
                numOfLanes = lanes;
                idealNumOfRacersTextField.setText("" +
                        (numOfLanes * (numOfLanes -1) * racesPerLane + 1));
            }
            catch(Exception err)  {
                numberOfLanesTextField.setText("" + numOfLanes);
            }
        }
    }
    private class RacesFocusListener implements FocusListener {

        public void focusGained(FocusEvent e) {
            // not used
        }

        public void focusLost(FocusEvent e) {
            int rpl;
            try  {
                rpl = Integer.parseInt(racesPerLaneTextField.getText());
                if(rpl < 1) throw new Exception();
                racesPerLane = rpl;
                idealNumOfRacersTextField.setText("" +
                        (numOfLanes * (numOfLanes -1) * racesPerLane + 1));
            }
            catch(Exception err)  {
                racesPerLaneTextField.setText("" + racesPerLane);
            }
        }
    }

    public void setRace(int lanes, int rpl)  {
        numberOfLanesTextField.requestFocus();
        numOfLanes = lanes;
        racesPerLane = rpl;
        numberOfLanesTextField.setText("" +lanes);
        racesPerLaneTextField.setText("" + rpl);
        calculateButton.requestFocus();
    }

    public int[][] getAvailableLineUp()  {
        int[][] temp = null;
        if (availableLineUp !=null)  {
            temp = new int [availableLineUp.length][availableLineUp[0].length];
            for (int row = 0; row < availableLineUp.length; row++)  {
                for (int col = 0; col < availableLineUp[0].length; col++)  {
                    temp[row][col] = availableLineUp[row][col]+1;
                }
            }
        }
        
        availableLineUp = null;
        return temp;
    }
}

